from invana_bot.utils.config import process_config


def process_pipe(parser_config=None):
    return process_config(parser_config)


def process_pipeline_config(crawlers=None):
    print("crawlers", crawlers)
    # TODO - validate the config too
    for crawler in crawlers:
        for extractor in crawler.get('parsers', []):
            extractor['data_selectors'] = process_pipe(extractor)['data_selectors']
    print("crawlers", crawlers)
    return crawlers



from scrapy.crawler import CrawlerProcess
from invana_bot.spiders.feeds import RSSSpider
import uuid

from scrapy.utils.log import configure_logging


class InvanaBotWebCrawlerBase(object):
    """


    """
    settings = {
        'COMPRESSION_ENABLED': False,  # this will save the data in normal text form, otherwise to bytes form.
        'HTTPCACHE_ENABLED': True,
        'TELNETCONSOLE_PORT': None

    }
    runner = None

    def __init__(self,
                 settings=None,
                 **kwargs):

        if settings is None:
            raise Exception("settings should be set")
        self.settings = settings
        self.job_id = self.generate_job_id()
        self.set_logger()

    @staticmethod
    def set_logger():
        configure_logging({'LOG_FORMAT': '%(levelname)s: %(message)s'})

    @staticmethod
    def generate_job_id():
        return uuid.uuid4().hex

    def _validate_urls(self, urls):
        if type(urls) is None:
            raise Exception("urls should be list type.")
        if len(urls) is 0:
            raise Exception("urls length should be atleast one.")

    def get_settings(self):
        return self.settings


"""
class InvanaFeedCrawler(InvanaBotWebCrawlerBase):

    def crawl_feeds(self, feed_urls=None):
        self.setup_crawler_type_settings(crawler_type="feeds")
        self._validate_urls(feed_urls)

        process = CrawlerProcess(self.settings)
        allowed_domains = []
        for feed_url in feed_urls:
            domain = feed_url.split("://")[1].split("/")[0]  # TODO - clean this
            allowed_domains.append(domain)

        process.crawl(RSSSpider,
                      start_urls=feed_urls,
                      )
        process.start()

"""



from .base import WebSpiderBase
from invana_bot.extractors.content import CustomContentExtractor, \
    ParagraphsExtractor, TableContentExtractor, HTMLMetaTagExtractor
import scrapy
from invana_bot.utils.url import get_domain, get_absolute_url
from invana_bot.utils.crawlers import get_crawler_from_list

TRAVERSAL_LINK_FROM_FIELD = "link_from_field"
TRAVERSAL_SAME_DOMAIN_FIELD = "same_domain"


class DefaultParserSpider(WebSpiderBase):
    """
    This is generic spider
    """
    name = "DefaultParserSpider"

    def closed(self, reason):
        print("spider closed with payload:", reason, self.current_crawler)

    @staticmethod
    def run_extractor(response=None, extractor=None):
        parser_name = extractor.get("parser_name")
        if parser_name in [None, "CustomContentExtractor"]:
            extractor_object = CustomContentExtractor(response=response, extractor=extractor)
        elif parser_name == "TableContentExtractor":
            extractor_object = TableContentExtractor(response=response, extractor=extractor)
        elif parser_name == "HTMLMetaTagExtractor":
            extractor_object = HTMLMetaTagExtractor(response=response, extractor=extractor)
        elif parser_name == "ParagraphsExtractor":
            extractor_object = ParagraphsExtractor(response=response, extractor=extractor)
        else:
            return
        data = extractor_object.run()
        return data

    @staticmethod
    def get_subdocument_key(parser=None, parser_name=None):
        """
        element is the subdocument key name.

        :param parser:
        :param parser_name:
        :return:
        """
        for extractor in parser['parsers']:
            if extractor.get("parser_name") == parser_name:
                for selector in extractor.get('data_selectors', []):
                    if selector.get('selector_attribute') == 'element':
                        return selector.get("id")
        return

    def parse(self, response=None):

        current_crawler = response.meta.get("current_crawler")
        crawlers = response.meta.get("crawlers")
        context = self.context

        if None in [crawlers, current_crawler]:
            current_crawler = self.current_crawler
            crawlers = self.crawlers

        data = {}
        for extractor in current_crawler['parsers']:
            extracted_data = self.run_extractor(response=response, extractor=extractor)
            data.update(extracted_data)
        if context is not None:
            data.update({"context": context})
        data['url'] = response.url
        data['domain'] = get_domain(response.url)
        yield data
        for traversal in current_crawler.get('traversals', []):
            if traversal['traversal_type'] == "pagination":
                # TODO - move this to run_pagination_traversal(self, response=None, traversal=None) method;
                traversal_config = traversal['pagination']
                max_pages = traversal_config.get("max_pages", 1)
                current_page_count = response.meta.get('current_page_count', 1)
                if current_page_count < max_pages:
                    next_selector = traversal_config.get('selector')
                    if next_selector:
                        if traversal_config.get('selector_type') == 'css':
                            next_page = response.css(next_selector + "::attr(href)").extract_first()
                        elif traversal_config.get('selector_type') == 'xpath':
                            next_page = response.xpath(next_selector + "::attr(href)").extract_first()
                        else:
                            next_page = None
                        current_page_count = current_page_count + 1
                        if next_page:
                            if not "://" in next_page:
                                next_page_url = "https://" + get_domain(response.url) + next_page
                            else:
                                next_page_url = next_page
                            # TODO - add logics to change the extractors or call a different parser from here.
                            yield scrapy.Request(
                                next_page_url,
                                callback=self.parse,
                                meta={"current_page_count": current_page_count}
                            )
            elif traversal['traversal_type'] == TRAVERSAL_LINK_FROM_FIELD:
                next_crawler_id = traversal['next_crawler_id']
                traversal_config = traversal[TRAVERSAL_LINK_FROM_FIELD]

                subdocument_key = self.get_subdocument_key(
                    parser=current_crawler,
                    parser_name=traversal_config['parser_name']
                )
                for item in data[subdocument_key]:
                    traversal_url = item[traversal[TRAVERSAL_LINK_FROM_FIELD]['field_name']]
                    next_parser = get_crawler_from_list(crawler_id=next_crawler_id, crawlers=crawlers)
                    yield scrapy.Request(
                        traversal_url, callback=self.parse,
                        meta={"crawlers": crawlers,
                              "current_crawler": next_parser
                              }
                    )
            elif traversal['traversal_type'] == TRAVERSAL_SAME_DOMAIN_FIELD:
                all_urls = response.css("a::attr(href)").extract()
                filtered_urls = []
                all_urls = list(set(all_urls))
                current_domain = get_domain(response.url)
                for url in all_urls:
                    url = get_absolute_url(url=url, origin_url=response.url)
                    if get_domain(url) == current_domain:
                        filtered_urls.append(url)
                filtered_urls = list(set(filtered_urls))
                # max_pages = traversal.get("max_pages", 100)
                #  implementing max_pages is difficult cos it keeps adding
                # new 100 pages in each thread.
                current_page_count = response.meta.get('current_page_count', 1)
                for url in filtered_urls:
                    current_page_count = current_page_count + 1

                    yield scrapy.Request(
                        url, callback=self.parse,
                        meta={"current_page_count": current_page_count}
                    )



"""
This component need to be revisited. Inspired by the component built for invana internal crawlers.
Need to find a way to implement with the current architecture.


"""


class ElementBase(object):
    nth_element = 0
    value_type = None

    def __init__(self, data_attribute=None, nth_element=None, **kwargs):
        if nth_element is not None:
            self.nth_element = nth_element
        self.data_attribute = data_attribute

    def validate_data(self):
        pass


class LinkElement(ElementBase):
    value_type = "href"


class HTMLElement(ElementBase):
    value_type = "innerHTML"


class TextElement(ElementBase):
    value_type = "text"


class ImageElement(ElementBase):
    value_type = "src"


class ListElement(object):
    is_multiple = True

    def __init__(self, instance, **kwargs):
        a = instance()
        pass



from invana_bot.extractors.base import ExtractorBase


class AllLinksExtractor(ExtractorBase):
    pass


class SameDomainLinkExtractor(ExtractorBase):
    pass


class PaginationLinkExtractor(ExtractorBase):
    pass


class CustomLinkExtractor(ExtractorBase):
    pass



from __future__ import unicode_literals
from pymongo import MongoClient
from transformers.executors import Executor
from datetime import datetime


class WriteToMongoDB(Executor):
    def __init__(self, connection_string, db, collection, unique_key, docs=None, update=True):
        self._connection_string = connection_string
        self._db = db
        self._collection = collection
        self._docs = docs
        self._unique_key = unique_key

    def connect(self):
        self._client = MongoClient(self._connection_string)

    def get_object(self, doc=None):
        return self._client[self._db][self._collection].find_one({self._unique_key: doc.get(self._unique_key)})

    def write(self):

        if isinstance(self._client, MongoClient):
            for doc in self._docs:
                entry = self.get_object(doc=doc)
                print("Writing the document ::{}".format(doc))
                if entry is None:
                    self._client[self._db][self._collection].insert(doc)
                else:
                    doc['updated'] = datetime.now()
                    self._client[self._db][self._collection].update_one({self._unique_key: doc.get(self._unique_key)},
                                                                        {"$set": doc})

    def disconnect(self):
        if isinstance(self._client, MongoClient):
            self._client.close()



from .base import InvanaBotWebCrawlerBase
from invana_bot.pipelines.default import CTIRunner
from datetime import datetime
from twisted.internet import reactor
import copy
from scrapy import signals
from scrapy.crawler import Crawler, CrawlerRunner
from scrapy.settings import Settings


class InvanaBotWebCrawler(InvanaBotWebCrawlerBase):
    """
    Invana bot web crawler

    """

    def create_job(self, cti_manifest=None, context=None):
        if context is None:
            context = {}
        if 'job_id' not in context.keys():
            context['job_id'] = self.job_id
            context['job_started'] = datetime.now()
        cti_runner = CTIRunner(cti_manifest=cti_manifest, settings=self.settings,
                               job_id=self.job_id, context=context)
        job, errors = cti_runner.crawl()
        return {"crawler_job": job, "crawler_job_errors": errors, "cti_runner": cti_runner}

    def start_job(self, job=None):
        runner = CrawlerRunner()
        crawler_job = job['crawler_job']
        cti_runner = job['cti_runner']
        spider_cls = crawler_job['spider_cls']
        spider_kwargs = crawler_job['spider_kwargs']

        def engine_stopped_callback():
            fn = copy.deepcopy(cti_runner.transform_and_index)
            fn()

        crawler = Crawler(spider_cls, Settings(cti_runner.settings))
        crawler.signals.connect(engine_stopped_callback, signals.engine_stopped)
        d = runner.crawl(crawler, **spider_kwargs)
        d.addBoth(engine_stopped_callback)
        reactor.run()



from invana_bot.spiders.default import DefaultParserSpider
from scrapy.linkextractors import LinkExtractor
from invana_bot.utils.crawlers import get_crawler_from_list
from invana_bot.utils.config import validate_cti_config
from scrapy.spiders import Rule
import copy
from pymongo import MongoClient
from transformers.transforms import OTManager
from transformers.executors import ReadFromMongo
from invana_bot.transformers.mongodb import WriteToMongoDB
from invana_bot.transformers.default import default_transformer
import requests


class ParserCrawler(object):
    """

    pipe_data = {  # single pipe

        "crawler_id": "blog-list",
        "parsers": [
            {
                "data_selectors": [
                    {
                        "id": "items",
                        "selector": ".post-listing .post-item",
                        "selector_attribute": "element",
                        "multiple": True
                    },
                    {
                        "id": "url",
                        "selector": ".post-header h2 a",
                        "selector_type": "css",
                        "selector_attribute": "href",
                        "parent_selector": "items",
                        "multiple": False
                    },
                    {
                        "id": "title",
                        "selector": ".post-header h2 a",
                        "selector_type": "css",
                        "selector_attribute": "text",
                        "parent_selector": "items",
                        "multiple": False
                    },
                    {
                        "id": "content",
                        "selector": ".post-content",
                        "selector_type": "css",
                        "selector_attribute": "html",
                        "parent_selector": "items",
                        "multiple": False
                    }
                ],
            }
        ],
        "traversals": [{
            "traversal_type": "pagination",
            "pagination": {
                "selector": ".next-posts-link",
                "selector_type": "css",
                "max_pages": 2
            },
        }]

    }

    context = {
        "client_id": "something"
    }

    """

    def __init__(self,
                 current_crawler=None,
                 start_urls=None,
                 job_id=None,
                 crawlers=None,
                 context=None):
        """

        :param current_crawler: single crawler in the CTI flow
        :param crawlers: all the crawlers in the CTI flow
        :param context: any extra information user want to send to the crawled data or carry forward.
        """
        self.current_crawler = current_crawler
        self.job_id = job_id
        self.crawlers = crawlers
        self.start_urls = start_urls
        if context:
            self.context = context
        self.validate_pipe()

    def validate_pipe(self):
        must_have_keys = ["crawler_id", "parsers"]
        optional_keys = ["traversals"]
        for key in must_have_keys:
            if key not in self.current_crawler.keys():
                raise Exception(
                    "invalid parser data, should have the following keys; {}".format(",".join(must_have_keys)))

    def validate_traversal(self):
        pass  # TODO - implement this

    def validate_extractor(self):
        pass  # TODO - implement this

    def get_traversals(self):
        return self.current_crawler.get("traversals", [])

    def get_extractors(self):
        return self.current_crawler.get("parsers", [])

    def generate_crawler_kwargs(self):
        domains = []

        for url in self.start_urls:
            domain = url.split("://")[1].split("/")[0]  # TODO - clean this
            domains.append(domain)

        extractor = LinkExtractor()
        rules = [
            Rule(extractor, follow=True)  # TODO - add regex types of needed.
        ]

        spider_kwargs = {
            "start_urls": self.start_urls,
            "allowed_domains": domains,
            "rules": rules,
            "current_crawler": self.current_crawler,
            "crawlers": self.crawlers,
            "context": self.context
        }
        return spider_kwargs

    def run(self):
        spider_cls = DefaultParserSpider
        spider_kwargs = self.generate_crawler_kwargs()
        return {"spider_cls": spider_cls, "spider_kwargs": spider_kwargs}


class CTIRunner(object):
    """


    """

    def __init__(self, cti_manifest=None, settings=None, job_id=None, context=None):
        self.cti_manifest = cti_manifest
        self.settings = settings
        self.crawlers = self.cti_manifest['crawlers']
        self.job_id = job_id
        self.context = context

    def run(self):
        return self.crawl()

    def crawl(self):
        errors = validate_cti_config(self.cti_manifest)
        if len(errors) == 0:

            initial_crawler = get_crawler_from_list(crawler_id=self.cti_manifest['init_crawler']['crawler_id'],
                                                    crawlers=self.crawlers)
            print("initial_crawler", initial_crawler)
            parser_crawler = ParserCrawler(
                job_id=self.job_id,
                start_urls=self.cti_manifest['init_crawler']['start_urls'],
                current_crawler=initial_crawler,
                crawlers=self.crawlers,
                context=self.context
            )
            cti_job = parser_crawler.run()

            return cti_job, errors
        else:
            return None, errors

    @staticmethod
    def get_index(transformation_id=None, indexes=None):
        for _index in indexes:
            if _index['transformation_id'] == transformation_id:
                return _index

    def index_data(self, index=None, results_cleaned=None):
        collection_name = index['collection_name']
        unique_key_field = index['unique_key']
        mongo_executor = WriteToMongoDB(
            self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['CONNECTION_URI'],
            self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['DATABASE_NAME'],
            collection_name,
            unique_key_field,
            docs=results_cleaned)
        mongo_executor.connect()
        mongo_executor.write()

    def trigger_callback(self, callback_config=None):
        callback_id = callback_config.get("callback_id")
        print("Triggering callback_id: {}".format(callback_id))

        url = callback_config.get("url")
        request_type = callback_config.get("request_type", '').lower()
        if request_type == "get":
            req = requests.get(url, headers=callback_config.get("headers", {}), verify=False)
            response = req.text
            print("Triggered callback successfully and callback responded with message :{}".format(response))
        elif request_type == "post":
            req = requests.post(url, json=callback_config.get("payload", {}),
                                headers=callback_config.get("headers", {}), verify=False)
            response = req.text
            print("Triggered callback successfully and callback responded with message :{}".format(response))

    def callback(self):
        all_indexes = self.cti_manifest.get('indexes', [])
        if len(all_indexes) == 0:
            print("There are no callback notifications associated with the indexing jobs. So we are Done here.")
        else:
            print("Initiating, sending the callback notifications after the respective transformations ")
            for index in all_indexes:
                index_id = index.get('index_id')
                callback_config = self.get_callback_for_index(index_id=index_id)
                try:
                    self.trigger_callback(callback_config=callback_config)
                except Exception as e:
                    print("Failed to send callback[{}] with error: {}".format(callback_config.get("callback_id"),
                                                                              e))

    def get_callback_for_index(self, index_id=None):
        callbacks = self.cti_manifest.get("callbacks", [])
        for callback in callbacks:
            callback_index_id = callback.get('index_id')
            if callback_index_id == index_id:
                return callback
        return

    def transform_and_index(self, callback=None):
        """
        This function will handle both tranform and index

        :param callback:
        :return:
        """
        print("transformer started")
        print("self.cti_manifest['transformations']", self.cti_manifest['transformations'])

        all_transformation = self.cti_manifest.get('transformations', [])
        if len(all_transformation) == 0:
            all_transformation.append({
                "transformation_id": "default"
            })

        for transformation in all_transformation:
            print("transformation", transformation)
            transformation_id = transformation['transformation_id']
            transformation_fn = transformation.get('transformation_fn')
            if transformation_fn is None:
                transformation_fn = default_transformer

            transformation_index_config = self.get_index(transformation_id=transformation_id,
                                                         indexes=self.cti_manifest['indexes'])
            mongo_executor = ReadFromMongo(
                self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['CONNECTION_URI'],
                self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['DATABASE_NAME'],
                self.settings['INVANA_BOT_SETTINGS']['ITEM_PIPELINES_SETTINGS']['COLLECTION_NAME'],
                query={"context.job_id": self.job_id}
            )
            mongo_executor.connect()
            ops = []
            ot_manager = OTManager(ops).process(mongo_executor)
            results = ot_manager.results
            results_cleaned__ = transformation_fn(results)
            results_cleaned = []
            for result in results_cleaned__:
                if "_id" in result.keys():
                    del result['_id']
                results_cleaned.append(result)
            self.index_data(index=transformation_index_config, results_cleaned=results_cleaned)

            print("Total results_cleaned count of job {} is {}".format(self.job_id, results.__len__()))

        print("======================================================")
        print("Successfully crawled + transformed + indexed the data.")
        print("======================================================")
        self.callback()



#!/usr/bin/env python

from setuptools import setup

setup(name='invana-bot',
      version='0.0.0c',
      description='A web crawler framework that can transform websites into datasets'
                  ' with Crawl, Transform and Index workflow.',
      author='Ravi Raja Merugu',
      author_email='rrmerugu@gmail.com',
      url='https://github.com/invanalabs/invana-bot',
      packages=[
          'tests',
          'invana_bot',
          'invana_bot.extractors',
          'invana_bot.fields',
          'invana_bot.httpcache',
          'invana_bot.pipelines',
          'invana_bot.spiders',
          'invana_bot.storages',
          'invana_bot.utils',
          'invana_bot.crawlers',
          'invana_bot.schedulers',
          'invana_bot.transformers',
          'invana_bot.managers'
      ],
      install_requires=[
          'Scrapy==1.6.0',
          'pymongo',
          'feedparser',
          'requests',
          'invana-transformers'
      ]
      )
